## OSM Geo-strip binning.
##
## Copyright © 2023,2024 Todd Coram, todd@maplefish.com, USA.

## Permission is hereby granted, free of charge, to any person obtaining
## a copy of this software and associated documentation files {the
## "Software"}, to deal in the Software without restriction, including
## without limitation the rights to use, copy, modify, merge, publish,
## distribute, sublicense, and/or sell copies of the Software, and to
## permit persons to whom the Software is furnished to do so, subject to
## the following conditions:

## The above copyright notice and this permission notice shall be
## included in all copies or substantial portions of the Software.

## THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
## EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
## MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
## NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
## LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
## OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
## WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

###############################################################################
## What is this??
##
## In a nutshell... we do binning of geographic points-of-interest based
## on strips of lattitude. For lack of a better name, this is
## strip-binning. The distance between latitudes, over the globe,  is
## constant, so we can set up our bins by simply specifying a constant
## width in minutes (or degrees if so inclined).

## We then use some Longitudal shortcuts to efficiently search through
## these strip bins looking for our points of interest.


## First thing's first
##

(scl 7)	## Minimum Resolution needed for haversine calculations


###############################################################################
## Some base constants calculations...
##
##

(load "@lib/math.l")

(setq *EARTH-CIRC-DEG 111.32)
(setq *EARTH-CIRC-MIN (/ *EARTH-CIRC-DEG 60))
(setq *width-minutes  0.5)

(de deg-to-rad (X)
    (*/ X (/ pi 180) 1.0))

(de geo-km-to-mile (K) (*/ 1.0 K 1.60934))
(de geo-mile-to-km (M) (*/ M 1.60934 1.0))
(de geo-km-to-min (K)
    (*/ 1.0 K *EARTH-CIRC-MIN))

###############################################################################
## Strip Binning Stuff
##

## geo-strips will primarily contain a property list indexed by "strips" of
## latitude, each strip is *width-minutes wide.
##
(setq *geo-strips '())

## Add some Data to a strip based on Latitude.  Data wil typically be
## (at least): Lat,Lon, a name and perhaps a physial address.
## For each strip identifies, the Data is concatenated with any other data
## that may already be there.
#
(de geo-add-to-strip (Lat Dat)
    (let (Sidx (/ Lat *width-minutes))
      (_geo-add-to-strip Sidx Dat)))

(de _geo-add-to-strip (idx val)
    (if (get '*geo-strips idx)
	(conc (get '*geo-strips idx) (list val))
      (put '*geo-strips idx (list val))))

## Load geolcation data from a file and bin them.
## This particular implementation uses a CSV file
## containing POI as data..
#
(de geo-strips-from-file (Csvf)
    (in Csvf
	(while (split (line) ",")
	  (let (Name (pack (get @  1))
    		House (pack (get @ 2))
		Street (pack (get @ 3))
		City (pack (get @ 4))
		State (pack (get @ 5))
		Lat (format (get @ 6) *Scl)
		Lon (format (get @ 7) *Scl)
		Cat (pack (get @ 8))
		Cusine (pack (get @ 9)))
	    (when (> (length Name) 0)
	      (geo-add-to-strip Lat
				(list Name House Street City State
				      Lat Lon Cat Cusine)))))) T)
  

###############################################################################
## Searching stuff.
##


## Standard Great Circle algorithm for find distance between two points.
## 
(de geo-haversine-dist (Th1 Ph1 Th2 Ph2)
    (setq
     Ph1 (*/ (- Ph1 Ph2) pi 180.0)
     Th1 (*/ Th1 pi 180.0)
     Th2 (*/ Th2 pi 180.0) )
    (let
	(DX (- (*/ (cos Ph1) (cos Th1) 1.0) (cos Th2))
            DY (*/ (sin Ph1) (cos Th1) 1.0)
            DZ (- (sin Th1) (sin Th2))
	    Km (* `(* 2 6371)
		  (asin
		   (/
		    (sqrt (+ (* DX DX) (* DY DY) (* DZ DZ)))
		    2))))
      Km))

## Find a range (list) of strips that are centered at the given Lattitude as
## defined by the given Km.
##
(de geo-find-strips (Lat Km)
    (let (Dist-min (geo-km-to-min Km)
	  My-idx (/ Lat *width-minutes)
	  Top-idx (- My-idx (/ (/ Dist-min 2) *width-minutes))
	  Bot-idx (+ My-idx (/ (/ Dist-min 2) *width-minutes)))
      (range Top-idx Bot-idx)))


## Calculate and return the longitudes Km in either direction (Lowe and Upper longitudes)
## from My-lon.
##
(de geo-hack-lon-range (My-lat My-lon Km)
    (let (Ldeg (lon-degs My-lat Km)
	  Lower (- My-lon Ldeg)
	  Upper (+ My-lon Ldeg))
      (list Lower Upper)))
(de len-1-deg-lon (Lat)
    (*/ (cos (deg-to-rad Lat)) *EARTH-CIRC-DEG 1.0))
(de lon-degs (Lat Km)
    (*/ Km 1.0 (len-1-deg-lon Lat)))

##
## Lookup to get a list (sorted by nearness) of all POI matching
## a passed category list. If the list is empty, match all categories.
## This function is optimized (via geo-hack-lon-range) to reduce the number
## of haversine distance calculations.
##
(de geo-nearby (My-lat My-lon Km Cats)
    (let (Idxs (geo-find-strips My-lat Km)
	  LoUp (geo-hack-lon-range My-lat My-lon Km)
	  Lower (car LoUp)
	  Upper (cadr LoUp)
	  Scnt 0
	  Distcnt 0
	  R NIL)
      (for Idx Idxs
	   (for O (get '*geo-strips Idx)
		(let (Name (get O 1)
			   Lat (get O 6)
			   Lon (get O 7)
			   Cat (get O 8)
			   SubCat (get O 9))
		  (inc 'Scnt)
		  (when (and (and (>= Lon Lower) (<= Lon Upper))
			     (or (= NIL Cats) (member Cat Cats)
				 (member SubCat Cats)))
		    (inc 'Distcnt)
		    (let (Okm (geo-haversine-dist My-lat My-lon Lat Lon))
		      (if (<= Okm Km)
			  (push 'R (append (list Okm) O))))))))
      (prinl (text
	      "Searching @1 bins (@2 items), haversine distancing @3 items --> @4 hits."
	      (length Idxs) Scnt Distcnt (length R)))
      (by 'car sort R)))

(de geo-init ()
    (wipe '*geo-strips))

## Converting from csv to property list is slow. This function allows saving
## the converted list so load time is dramatically reduced. Must be called
## from the REPL...
#
(de geo-csv-to-native (Fins Fout)
    (when (atom Fins)
	(setq Fins (list Fins)))
    (geo-init)
    (for Fin Fins
	 (msg "Loading: " Fin)
	 (geo-strips-from-file Fin))
    (out Fout
	 (print (getl '*geo-strips))) T)

## Loading native format is super fast.
##
(de geo-load-native (Fin)
    (putl '*geo-strips (in Fin (read))) T)

(de geo-test1 ()
    (geo-init)
    (msg "load")
    (bench (putl '*geo-strips (in "conus.dat" (read))))
    (msg "run")
    (test-nearby 10.0))

(de test-nearby (Km)
    (gc)
    (bench (msg (length (geo-nearby 38.910844556948106, -77.04372018213103 (geo-mile-to-km Km))))))

